"""
About DChemPy:
    Made by Yiheng Dai
    Last Update: 2022.11.21
    Most codes were completed during 2020-2022 in Xiamen University.
    Made for Machine Learning projects without deep learning in chemistry.

Contents:
    DChemPy.Chem: Process SMILES
    DChemPy.Dataset: Modify Dataset
    DChemPy.Drawing: Draw Plots in Machine Learning
    DChemPy.Featurization: Make SMILES into Feature Matrix
    DChemPy.Statistic: Basic Statistic Analysis

Original Paper:
    In press...
"""

from .Chem import *
from .Dataset import *
from .Drawing import *
from .Statistic import *
from .Featurization import *

VERSION = '1.1.1'
